/*
 * Configuration.cs
 * $Id: Configuration.cs,v 1.7 2004/03/31 08:04:56 Mayuki Sawatari Exp $
 *
 * Copyright (C) 2003 Mayuki Sawatari, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE MISUZILLA.ORG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
 
using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;
using System.Resources;
using System.ComponentModel;
using System.Collections;
using System.Collections.Specialized;
using System.Runtime.InteropServices;

using Misuzilla.Utilities;
using Misuzilla.Windows.Forms;
using Misuzilla.Net.IRC.Ircv;

namespace
Misuzilla.Net.IRC.Ircv
{
	internal class
	Configuration
	{
		private IRCLogViewer _ircv = null;
		
		private String _server  = "localhost";
		private Int32  _port    = 6668;
		private String _user    = "user";
		private String _nick    = "nick";
		private String _passwd  = "";
		private String _realname  = "realname";
		private String _charset   = "ISO-2022-JP";
		
		private Font _textFont;
		private Boolean _useInputForm = false;
		private Boolean _shadow       = false;
		private Double _FadeWait = 25;
		private Byte _FadeRatio  = 25;
		private Byte _logFormOpacity    = 255;
		private Byte _logFormOpacityMin = 77;
		private Byte _inputFormOpacity  = 255;
		private Int32 _stringHeight = 0;
		private Int32 _logLines = 10;
		private Int32 _width    = 0;
		private Int32 _left     = 0;
		private Int32 _top      = 0;
		private Int32 _inputFormTop     = 0;
		private Int32 _inputFormLeft    = 0;
		private Int32 _tonedownInterval = 5000;
		
		private Regex _ignore = null;
		private Regex _keyword = null;
		private String _browserPath = "";
		private String _networkSeparator = "@";
		private String[] _inputFormNumAccessChannels = null;
		private StringCollection _ignoreChannels     = null;
		private MessageTypes _visibleMessageType = MessageTypes.PrivMsg;
		
		private HotkeyModifiers _hotkeyMod         = HotkeyModifiers.None;
		private HotkeyModifiers _hotkeyModFade     = HotkeyModifiers.None;
		private HotkeyModifiers _hotkeyModInputWnd = HotkeyModifiers.None;
		private Keys _hotkeyKey                = Keys.None;
		private Keys _hotkeyKeyFade            = Keys.None;
		private Keys _hotkeyKeyInputWnd        = Keys.None;
		private Keys _channelUpDownModifierKey = Keys.Control;
	
		private StringDictionary _logFormat = null;
		//private Char   _logFormatNetworkSeparator = '@';
		
		public
		Configuration(IRCLogViewer ircv)
		{
			_logFormat = new StringDictionary();
			_ircv = ircv;
		}
		public void
		LoadConfigurations(String path)
		{
			Rectangle rectscreen = Screen.GetWorkingArea(_ircv);
			StringDictionary sd = Setting.GetSettings(path);

			// IRC
			_port = StringToInt32(sd["irc.port"], 6667); // 6667
			_server = sd["irc.server"];
			_nick = sd["irc.nick"];
			_passwd = sd["irc.passwd"];
			_user = sd["irc.user"];
			if (sd["irc.realname"] != null) _realname = sd["irc.realname"];
			if (sd["irc.charset"] != null) _charset = sd["irc.charset"];
			if (sd["irc.networkseparator"] != null) _networkSeparator = sd["irc.networkseparator"];

			_tonedownInterval = StringToInt32(sd["window.tonedowninterval"], 5000); // 10
			_logLines = StringToInt32(sd["window.loglines"], 10); // 10
			_shadow = (String.Compare(sd["window.shadow"], "true", true) == 0);
			
			if (String.Compare(sd["window.visiblejoinpart"], "true", true) == 0)
				_visibleMessageType |= MessageTypes.Join | MessageTypes.Part;

			if (sd["window.fade.ratio"] != null) _FadeRatio = Byte.Parse(sd["window.fade.ratio"]);
			if (sd["window.fade.wait"] != null) _FadeWait = Double.Parse(sd["window.fade.wait"]);
			if (sd["window.opacity"] != null) _logFormOpacity = Byte.Parse(sd["window.opacity"]);
			if (sd["window.input.opacity"] != null) _inputFormOpacity = Byte.Parse(sd["window.input.opacity"]);
			if (sd["window.opacity.minimum"] != null)
				_logFormOpacityMin = Byte.Parse(sd["window.opacity.minimum"]);
			else
				_logFormOpacityMin = (Byte)(_logFormOpacity * 0.3);

			if (sd["window.keyword"] != null) _keyword = new Regex(sd["window.keyword"]);
			if (sd["window.browser"] != null) _browserPath = sd["window.browser"];


			// font
			String fontFamily = "Tahoma";
			Single fontSize = 9;
			if (sd["window.font.family"] != null)
				fontFamily = sd["window.font.family"];
			if (sd["window.font.size"] != null)
				fontSize = Single.Parse(sd["window.font.size"]);
			_ircv.Font = _textFont = new System.Drawing.Font(fontFamily, fontSize, FontStyle.Bold);
			_stringHeight = (Int32)_ircv.Graphics.MeasureString("|", _textFont).Height + 2;
			
			// messagetype
			if (sd["window.messagetype"] != null)
				//「,」区切りなら Enum.Parse(s, typeof(MessageTypes), true); で済む(←バカ。
				foreach (String s in sd["window.messagetype"].Split(' ')) {
					switch (s.ToUpper()) {
					case "NOTICE":
						_visibleMessageType |= MessageTypes.Notice;
						break;
					case "NICK":
						_visibleMessageType |= MessageTypes.Nick;
						break;
					case "JOIN":
						_visibleMessageType |= MessageTypes.Join;
						break;
					case "PART":
						_visibleMessageType |= MessageTypes.Part;
						break;
					case "QUIT":
						_visibleMessageType |= MessageTypes.Quit;
						break;
					case "TOPIC":
						_visibleMessageType |= MessageTypes.Topic;
						break;
					case "NUM":
						_visibleMessageType |= MessageTypes.NumericReply;
						break;
					}
				}
			
			Console.WriteLine(_visibleMessageType);
			
			_width = StringToInt32(sd["window.width"], rectscreen.Width / 2);
			_left = StringToInt32(sd["window.left"], rectscreen.Left + _width);
			_top = StringToInt32(sd["window.top"], (rectscreen.Height - (_stringHeight * _logLines)) - 2);
			
			_inputFormLeft = StringToInt32(sd["window.inputform.left"], _left);
			_inputFormTop = StringToInt32(sd["window.inputform.top"], _top);

			// Hotkey
			// スタイルが変わるとHandleが変わる。
			_hotkeyKeyInputWnd = _hotkeyKeyFade = _hotkeyKey = Keys.None;
			_hotkeyModInputWnd = _hotkeyModFade = _hotkeyMod = HotkeyModifiers.None;
			if (sd["window.visible.hotkey.modifiers"] != null && sd["window.visible.hotkey.key"] != null) {
				try {
					_hotkeyMod = (HotkeyModifiers)Enum.Parse(typeof(HotkeyModifiers), sd["window.visible.hotkey.modifiers"], true);
					_hotkeyKey = (Keys)Enum.Parse(typeof(Keys), sd["window.visible.hotkey.key"], true);
				} catch (System.ArgumentException) {
					// 不正
				}
			}
			if (sd["window.fade.hotkey.modifiers"] != null && sd["window.fade.hotkey.key"] != null) {
				try {
					_hotkeyModFade = (HotkeyModifiers)Enum.Parse(typeof(HotkeyModifiers), sd["window.fade.hotkey.modifiers"], true);
					_hotkeyKeyFade = (Keys)Enum.Parse(typeof(Keys), sd["window.fade.hotkey.key"], true);
				} catch (System.ArgumentException) {
					// 不正
				}
			}
			if (sd["window.inputform.hotkey.modifiers"] != null && sd["window.inputform.hotkey.key"] != null) {
				try {
					_hotkeyModInputWnd = (HotkeyModifiers)Enum.Parse(typeof(HotkeyModifiers), sd["window.inputform.hotkey.modifiers"], true);
					_hotkeyKeyInputWnd = (Keys)Enum.Parse(typeof(Keys), sd["window.inputform.hotkey.key"], true);
				} catch (System.ArgumentException) {
					// 不正
				}
			}
			
			//
			if (sd["window.inputform.channelupdownselect.modifiers"] != null) {
				try {
					_channelUpDownModifierKey = (Keys)Enum.Parse(typeof(Keys), sd["window.inputform.channelupdownselect.modifiers"], true);
				} catch (System.ArgumentException) {
					// 不正
				}
			}
			
			// channel
			if (sd["window.ignorechannels"] != null) {
				_ignoreChannels = new StringCollection();
				foreach (String channel in sd["window.ignorechannels"].Split(new Char[] { ' ' } )) {
					// StringDictionaryは case sensitive なので手抜き。
					_ignoreChannels.Add(channel.ToLower());
				}
			}
			// ignore
			if (sd["window.ignoresenders"] != null) {
				StringBuilder sb = new StringBuilder();
				sb.Append("^(?:");
				foreach (String s in sd["window.ignoresenders"].Split(new Char[] {' '})) {
					if (s != "") {
						sb.Append(Regex.Escape(s)).Append("|");
					}
				}
				if (sb.Length != 4) {
					sb.Remove(sb.Length-1, 1); // 余分な | を消す。
					sb.Append(")$").Replace("\\*", ".*");
					_ignore = new Regex(sb.ToString(), RegexOptions.IgnoreCase);
				} else {
					_ignore = null;
				}
			}
			// accesskey
			_inputFormNumAccessChannels = new String[10];
			for (Int32 i = 0; i < 10; i++) {
				if (sd["window.inputform.channelaccesskey."+i.ToString()] != null) {
					_inputFormNumAccessChannels[i] = sd["window.inputform.channelaccesskey."+i.ToString()];
				}
			}
			
			// inputform
			if (sd["window.inputform.use"] != null) {
				_useInputForm = Boolean.Parse(sd["window.inputform.use"]);
			}
			// 以前のバージョンで使っていたもの
			if (sd["-.window.inputform.use"] != null) {
				_useInputForm = Boolean.Parse(sd["-.window.inputform.use"]);
			}
			
			// format
			if (sd["window.logformat.privmsg"] != null)
				_logFormat["privmsg"] = sd["window.logformat.privmsg"];
			else
				_logFormat["privmsg"] = "{receiver_wo_network}/{sender}: {message}";
					
			if (sd["window.logformat.other"] != null)
				_logFormat["other"] = sd["window.logformat.other"];
			else
				_logFormat["other"] = "{message}";
		}
		
		//
		// properties
		//
		[Category("ホットキー"), Description("同時に押される特殊なキーを指定します。")]
		public HotkeyModifiers
		LogWindowHotkeyModifier
		{
			get { return _hotkeyMod; }
			set { _hotkeyMod = value; }
		}
		[Category("ホットキー"), Description("同時に押される特殊なキーを指定します。")]
		public HotkeyModifiers
		InputWindowHotkeyModifier
		{
			get { return _hotkeyModInputWnd; }
			set { _hotkeyModInputWnd = value; }
		}
		[Category("ホットキー"), Description("同時に押される特殊なキーを指定します。")]
		public HotkeyModifiers
		LogWindowFadeHotkeyModifier
		{
			get { return _hotkeyModFade; }
			set { _hotkeyModFade = value; }
		}
			
		[Category("ホットキー"), Description("メニューを表示するためのキーを指定します。")]
		public Keys
		LogWindowHotkey
		{
			get { return _hotkeyKey; }
			set { _hotkeyKey = value; }
		}
		[Category("ホットキー"), Description("発言フォームを表示するためのキーを指定します。")]
		public Keys
		InputWindowHotkey
		{
			get { return _hotkeyKeyInputWnd; }
			set { _hotkeyKeyInputWnd = value; }
		}
		[Category("ホットキー"), Description("ログのフェード状態を切り替えるためのキーを指定します。")]
		public Keys
		LogWindowFadeHotkey
		{
			get { return _hotkeyKeyFade; }
			set { _hotkeyKeyFade = value; }
		}
		public Keys
		ChannelUpDownModifierKey
		{
			get { return _channelUpDownModifierKey; }
			set { _channelUpDownModifierKey = value; }
		}
		
		public Double
		FadeWait
		{
			get { return _FadeWait; }
			set { _FadeWait = value; }
		}
		public Byte
		FadeRatio
		{
			get { return _FadeRatio; }
			set { _FadeRatio = value; }
		}
		public Byte
		LogWindowOpacity
		{
			get { return _logFormOpacity; }
			set { _logFormOpacity = value; }
		}
		public Byte
		LogWindowOpacityMinimum
		{
			get { return _logFormOpacityMin; }
			set { _logFormOpacityMin = value; }
		}
		public Byte
		InputWindowOpacity
		{
			get { return _inputFormOpacity; }
			set { _inputFormOpacity = value; }
		}
		
		public Boolean
		Shadow
		{
			get { return _shadow; }
			set { _shadow = value; }
		}
		public Boolean
		UseInputForm
		{
			get { return _useInputForm; }
			set { _useInputForm = value; }
		}
		public Int32
		TonedownInterval
		{
			get { return _tonedownInterval; }
			set { _tonedownInterval = value; }
		}
		public Int32
		StringHeight
		{
			get { return _stringHeight; }
			set { _stringHeight = value; }
		}
		public Int32
		LogLines
		{
			get { return _logLines; }
			set { _logLines = value; }
		}
		public Int32
		Top
		{
			get { return _top; }
			set { _top = value; }
		}
		public Int32
		Left
		{
			get { return _left; }
			set { _left = value; }
		}
		public Int32
		Width
		{
			get { return _width; }
			set { _width = value; }
		}
		public Int32
		InputFormTop
		{
			get { return _inputFormTop; }
			set { _inputFormTop = value; }
		}
		public Int32
		InputFormLeft
		{
			get { return _inputFormLeft; }
			set { _inputFormLeft = value; }
		}
		public MessageTypes
		VisibleMessageType
		{
			get { return _visibleMessageType; }
			set { _visibleMessageType = value; }
		}
		public StringDictionary
		LogFormat
		{
			get { return _logFormat; }
			set { _logFormat = value; }
		}
		public String
		BrowserPath
		{
			get { return _browserPath; }
			set { _browserPath = value; }
		}
		public Regex
		Ignore
		{
			get { return _ignore; }
			set { _ignore = value; }
		}
		public Regex
		Keyword
		{
			get { return _keyword; }
			set { _keyword = value; }
		}
		public String[]
		InputFormNumAccessChannels
		{
			get { return _inputFormNumAccessChannels; }
			set { _inputFormNumAccessChannels = value; }
		}
		public StringCollection
		IgnoreChannels
		{
			get { return _ignoreChannels; }
			set { _ignoreChannels = value; }
		}
			
		[Category("ネットワーク接続")]
		public String
		Server
		{
			get { return _server; }
			set { _server = value; }
		}
		[Category("ネットワーク接続")]
		public String
		Password
		{
			get { return _passwd; }
			set { _passwd = value; }
		}
		[Category("ネットワーク接続")]
		public String
		User
		{
			get { return _user; }
			set { _user = value; }
		}
		[Category("ネットワーク接続")]
		public String
		RealName
		{
			get { return _realname; }
			set { _realname = value; }
		}
		[Category("ネットワーク接続")]
		public String
		Charset
		{
			get { return _charset; }
			set { _charset = value; }
		}
		[Category("ネットワーク接続")]
		public String
		Nick
		{
			get { return _nick; }
			set { _nick = value; }
		}
		[Category("ネットワーク接続")]
		public Int32
		Port
		{
			get { return _port; }
			set { _port = value; }
		}
		[Category("ネットワーク接続")]
		public String
		ChannelNetworkSeparator
		{
			get { return _networkSeparator; }
			set { _networkSeparator = value; }
		}		
		//
		private static Int32
		StringToInt32(String s, Int32 dvalue)
		{
			try {
				return Int32.Parse(s);
			} catch (ArgumentNullException) {
				return dvalue;
			} catch (FormatException) {
				return dvalue;
			} // throw
		}
	}
}