/*
 * IRCClient.cs
 * $Id: IRCClient.cs,v 1.12 2004/02/19 13:30:08 Mayuki Sawatari Exp $
 *
 * Copyright (C) 2003-2004 Mayuki Sawatari, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE MISUZILLA.ORG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
 
using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;
using System.Resources;
using System.Collections;
using System.Collections.Specialized;
using System.Diagnostics;

namespace
Misuzilla.Net.IRC
{
	// IRCを管理するクラス。一クラス/一スレッド/一サーバー
	public delegate void IRCMessageDelegate(IRCMessage im);
	public class
	IRCClient// : IDisposable
	{
		private TcpClient _tcpclient;

		private String  _user;
		private String  _passwd;
		private String  _nick;
		private String  _realname;
		private String  _charset;
		private Int32   _port;
		private String  _server;
		private Boolean _running = false;
		private Boolean _connected = false;
		private Int32   _retryInterval = 15000;
		
		private Thread _thread;
		private NetworkStream _nstream;
		private StreamReader _sr;
		private StreamWriter _sw;
		
		public event IRCMessageDelegate OnMessageReceived;
		public String CurrentRawMessage = "";
		
		public Boolean
		Running
		{
			get { return _running; }
		}
		public Boolean
		Connected
		{
			get { return _connected; }
		}
		
		public
		IRCClient(String server, Int32 port, String user, String passwd, String nick, String charset)
			: this(server, port, user, passwd, nick, "", charset)
		{
		}
		public
		IRCClient(String server, Int32 port, String user, String passwd, String nick, String realname, String charset) 
		{
			_server = server;
			_port = port;
			_user = user;
			_passwd = passwd;
			_nick = nick;
			_realname = realname;
			_charset = charset;
			
			_thread = new Thread(new ThreadStart(this.Run)); 
		}

		public void
		Start() 
		{ 
			this.Connect();
			_thread.Start();
		} 

		public void
		Run()
		{
			while (_running) {
				while (true) {
					try {
						String s = _sr.ReadLine();
						if (s == null) break;
						this.CurrentRawMessage = s;
						OnMessageReceived(Parse(s));
						Thread.Sleep(1);
					} catch (NullReferenceException) {
						break;
					} catch (IOException) {
						break;
					} catch (IRCMessageInvalidException) {
					} catch (ThreadAbortException) {
						break;
					}
				}
				//Trace.WriteLine(ie.ToString());
				_connected = false;
				OnMessageReceived(new NoticeMessage(":Ircv!~ircv@localhost NOTICE Ircv :切断されました"));

				if (_retryInterval > 0 && _running && !_connected) {
					OnMessageReceived(new NoticeMessage(":Ircv!~ircv@localhost NOTICE Ircv :再接続しています..."));
					// reconnect
					try {
						this.Connect();
					} catch (SocketException e) {
						OnMessageReceived(new NoticeMessage(":Ircv!~ircv@localhost NOTICE Ircv :" + e.Message));
						Thread.Sleep(_retryInterval);
					}
				} else {
					_running = false;
				}
			}
		}
		
		public void
		Connect()
		{
			if (_sw != null) { _sw.Close(); _sw = null; }
			if (_sr != null) { _sr.Close(); _sr = null; }
			if (_tcpclient != null) { _tcpclient.Close(); _tcpclient = null; }
			
			_tcpclient = new TcpClient(_server, _port);
			_nstream = _tcpclient.GetStream();

			if (_charset.ToLower() != "utf-8") {
				_sr = new StreamReader(_nstream, Encoding.GetEncoding(_charset));
			} else {
				_sr = new StreamReader(_nstream);
			}
			if (_charset.ToLower() != "utf-8") {
				_sw = new StreamWriter(_nstream, Encoding.GetEncoding(_charset));
			} else {
				_sw = new StreamWriter(_nstream);
			}
			
			_running = true;

			this.OnMessageReceived += new IRCMessageDelegate(MessageReceived);

			// nick
			this.Send("NICK " + _nick);
			// pass
			if (_passwd.Length != 0) this.Send("PASS " + _passwd);
			// user
			this.Send("USER " + _user + " * * :" + _realname);

			_connected = true;
			
		}
		
		public void
		Disconnect()
		{
			try {
				if (_tcpclient != null && _sw != null)
					this.Send(new QuitMessage());
			} catch (IOException) {
			}

			_running = false;
			if (_sw != null) { _sw.Close(); _sw = null; }
			if (_sr != null) { _sr.Close(); _sr = null; }
			if (_tcpclient != null) { _tcpclient.Close(); _tcpclient = null; }

			if (_thread.ThreadState == System.Threading.ThreadState.Running) {
				_thread.Join(1000);
				_thread.Abort();
			}
		}
		
		public void
		Send(String m)
		{
			Trace.WriteLine("Send: " + m);
			_sw.WriteLine(m);
			_sw.Flush();
		}
		public void
		Send(IRCMessage m)
		{
			this.Send(m.RawMessage);
		}
		
		private IRCMessage
		Parse(String s)
		{
			IRCMessage im = IRCMessage.CreateMessage(s);
			im.IsServerMessage = true;
			return im;
		}
		
		private void
		MessageReceived(IRCMessage im)
		{
			//Trace.WriteLine(im);
		}
		
		
		/* CTCPを見つけるとdelegateを呼んでそれの戻り値をつかって文字列にする */
		public delegate String FoundCTCPCommand(IRCMessage im, String command);
		public static IRCMessage
		ParseCTCP(IRCMessage pm, FoundCTCPCommand f)
		{
			if (pm.Content.IndexOf('\u0001') < 0) return pm;
			
			StringBuilder sb = new StringBuilder();
			Boolean inCommand = false;
			StringBuilder sbCmd = new StringBuilder();
			for (Int32 i = 0; i < pm.Content.Length; i++) {
				Char c = pm.Content[i];
				if (c == '\u0001') {
					if (inCommand) {
						if (f != null) {
							sb.Append(f(pm, sbCmd.ToString()));
							sbCmd = new StringBuilder();
						}
					}
					inCommand = !inCommand;
				} else {
					if (inCommand)
						sbCmd.Append(c);
					else
						sb.Append(c);
				}
			}
			pm.Content = sb.ToString();
			
			return pm;
		}

		[Obsolete]
		public String
		ParseCTCP(IRCMessage im, String msg, FoundCTCPCommand f)
		{
			if (msg.IndexOf('\u0001') < 0) return msg;
			
			StringBuilder sb = new StringBuilder();
			Boolean inCommand = false;
			StringBuilder sbCmd = new StringBuilder();
			for (Int32 i = 0; i < msg.Length; i++) {
				Char c = msg[i];
				if (c == '\u0001') {
					if (inCommand) {
						if (f != null) {
							sb.Append(f(im, sbCmd.ToString()));
							sbCmd = new StringBuilder();
						}
					}
					inCommand = !inCommand;
				} else {
					if (inCommand)
						sbCmd.Append(c);
					else
						sb.Append(c);
				}
			}
			
			return sb.ToString();
		}
	}

	public abstract class 
	IRCMessage
	{
		protected Boolean _isServerMessage = false;
		protected String _rawMessage = "";
		protected String _content = "";
		protected String _senderHost = "";
		protected String _senderNick = "";
		protected String _sender = "";
		protected String _prefix = "";
		protected String _command = "";
		protected String _commandParam = "";
		protected String[] _commandParams;
		
		private void
		SetPrefix()
		{
			if (_senderNick == "")
				_prefix = "";
			else if (_senderHost == "")
				_prefix = _senderNick;
			else
				_prefix = _senderNick + "!" + _senderHost;
			
		}
		public Boolean IsServerMessage
		{
			get { return _isServerMessage; }
			set { _isServerMessage = value; }
		}
		public String Sender
		{
			get { return _sender; }
			set
			{
				String[] part = value.Split(new char[] {'!'}, 2);
				if (part.Length == 2) {
					// nick!~user@host
					_senderNick = part[0];
					_senderHost = part[1];
				} else {
					// nick
					_senderHost = "";
					_senderNick = value;
				}
				_sender = value;
				this.SetPrefix();
			}
		}
		public String SenderHost
		{
			get { return _senderHost; }
			set { _senderHost = value; this.SetPrefix(); }
		}
		public String SenderNick
		{
			get { return _senderNick; }
			set { _senderNick = value; this.SetPrefix(); }
		}
		
		public String Prefix
		{
			get { return _prefix; }
			set { this.Sender = value; }
		}
		public String Command
		{
			get { return _command; }
		}
		public String CommandParam
		{
			get {
				StringBuilder sb = new StringBuilder();
				for (Int32 i = 0; i < _commandParams.Length; i++) {
					if (i+1 == _commandParams.Length || _commandParams[i+1] == null || _commandParams[i+1] == "") {
						sb.Append(":");
						sb.Append(_commandParams[i]);
						sb.Append(" ");
						break;
					} else {
						sb.Append(_commandParams[i]);
						sb.Append(" ");
					}
				}
				sb.Remove(sb.Length-1, 1);
				_commandParam = sb.ToString();

				//return _commandParam;
				return sb.ToString();
			}
			set {
				String[] param = value.Split(new Char[]{' '});
				_commandParam = value;
				_commandParams = new String[15];
				for (Int32 i = 0; i < 15; i++) {
					if (i < param.Length && i < _commandParams.Length) {
						if (param[i].StartsWith(":")) {
							_commandParams[i] = String.Join(" ", param, i, param.Length - i).Substring(1);
							break;
						} else {
							_commandParams[i] = param[i];
						}
					} else {
						_commandParams[i] = "";
					}
				}
			}
		}
		public String[] CommandParams
		{
			get { return _commandParams; }
			set { _commandParams = value; }
		}
		
		public String RawMessage
		{
			get {
				StringBuilder sb = new StringBuilder();
				if (_prefix.Length != 0) {
					sb.Append(":");
					sb.Append(_prefix);
					sb.Append(" ");
				}
				sb.Append(_command);
				sb.Append(" ");
				sb.Append(this.CommandParam);
				
				//return _rawMessage;
				return sb.ToString();
			}
		}
		public virtual String Content
		{
			get { return _content; }
			set { _content = value; }
		}

		public new virtual String
		ToString()
		{
			return RawMessage;
		}

		public
		IRCMessage(String raw)
		{
			String[] parts;
			_rawMessage = raw;
			
			try {
				if (raw.StartsWith(":")) {
					parts = raw.Split(new Char[]{' '}, 3);
					_prefix = parts[0].Substring(1);
					_command = parts[1];
					_commandParam = parts[2];
				} else {
					parts = raw.Split(new Char[]{' '}, 2);
					_command = parts[0];
					_commandParam = parts[1];
					_prefix = "";
				}
			} catch (IndexOutOfRangeException) {
				throw new IRCMessageInvalidException("メッセージの形式が不正です\nメッセージ: " + raw);
			}
			
			this.CommandParam = _commandParam;
			this.Sender = _prefix;
			
		}
		public
		IRCMessage()
		{
			_commandParams = new String[15];
		}
		
		public static IRCMessage
		CreateMessage(String raw)
		{
			Trace.WriteLine(raw);
			OtherMessage om = new OtherMessage(raw);
			
			switch (om.Command.ToUpper()) {
			case "NOTICE":
				return new NoticeMessage(raw);
			case "PRIVMSG":
				return new PrivMsgMessage(raw);
			case "NICK":
				return new NickMessage(raw);
			case "JOIN":
				return new JoinMessage(raw);
			case "PART":
				return new PartMessage(raw);
			case "QUIT":
				return new QuitMessage(raw);
			case "TOPIC":
				return new TopicMessage(raw);
			default: //case "372": case "001": case "002":
				try {
					if (Int32.Parse(om.Command) > 0)
						return new NumericReplyMessage(raw);
				} catch (FormatException) {
				}
				return om;
			}
		}
	}

	public class
	OtherMessage : IRCMessage
	{
		public
		OtherMessage(String raw) : base(raw)
		{
		}
		public override String
		ToString()
		{
			return _rawMessage;
		}
	}

	public class
	NumericReplyMessage : IRCMessage
	{
		private Int32 _replyNumber = 0;
		
		public
		NumericReplyMessage(Int32 i)
		{
			_replyNumber = i;
			_command = i.ToString("000");
		}
		public
		NumericReplyMessage(String raw) : base(raw)
		{
			_replyNumber = Int32.Parse(_command);
			_content = _commandParams[1];
		}
		
		public Int32
		ReplyNumber
		{
			get { return _replyNumber; }
			set { _replyNumber = value; }
		}
		
		public override String
		ToString()
		{
			return _replyNumber.ToString("000") + " : " + _commandParams[1];
			//return _rawMessage;
		}
	}
	
	public class
	PrivMsgMessage : IRCMessage
	{
		private String _receiver;
		
		public String Receiver
		{
			get { return _receiver; }
			set { _commandParams[0] = _receiver = value; }
		}
		public override String Content
		{
			set { _commandParams[1] = _content = value; }
		}

		public
		PrivMsgMessage() : base()
		{
			_command = "PRIVMSG";
		}
		public
		PrivMsgMessage(String raw) : base(raw)
		{
			_command = "PRIVMSG";
			_receiver = _commandParams[0];
			_content = _commandParams[1];
		}
		public
		PrivMsgMessage(String receiver, String message) : base()
		{
			_command = "PRIVMSG";
			this.Receiver = receiver;
			this.Content = message;
		}

		public override String
		ToString()
		{
			return String.Format("PrivMsg: ({0}->{1}) {2}", _senderNick, _receiver, _content);
		}
	}
	public class
	NickMessage : IRCMessage
	{
		private String _newNick;
		public String NewNick
		{
			get { return _newNick; }
			set { _commandParams[0] = _newNick = value; }
		}
		
		public
		NickMessage() : base()
		{
			_command = "NICK";
		}
		public
		NickMessage(String raw) : base(raw)
		{
			_command = "NICK";
			_newNick = _commandParams[0];
		}

		public override String
		ToString()
		{
			return String.Format("Nick: {0} -> {1}", _senderNick, _newNick);
		}
	}
	public class
	NoticeMessage : IRCMessage
	{
		private String _receiver;
		public String Receiver
		{
			get { return _receiver; }
			set { _commandParams[0] = _receiver = value; }
		}
		public override String Content
		{
			set { _commandParams[1] = _content = value; }
		}
		
		public
		NoticeMessage() : base()
		{
			_command = "NOTICE";
		}
		public
		NoticeMessage(String raw) : base(raw)
		{
			_command = "NOTICE";
			_receiver = _commandParams[0];
			_content = _commandParams[1];
		}
		public
		NoticeMessage(String receiver, String message) : base()
		{
			_command = "NOTICE";
			this.Receiver = receiver;
			this.Content = message;
		}

		public override String
		ToString()
		{
			if (this.Sender.Length != 0)
				return String.Format("Notice: ({0}) [{1}] {2}", _senderNick, _receiver, _content);
			else 
				return String.Format("Notice: [{0}] {1}",  _receiver, _content);
		}
	}
	public class
	JoinMessage : IRCMessage
	{
		private String _channel;
		public String Channel
		{
			get { return _channel; }
			set { _channel = value; }
		}
		
		public
		JoinMessage() : base()
		{
			_command = "JOIN";
		}
		public
		JoinMessage(String raw) : base(raw)
		{
			_channel = _commandParams[0];
			_command = "JOIN";
		}
		public
		JoinMessage(String channel, String param)
		{
			_command = "JOIN";
			_channel = channel;
		}

		public override String
		ToString()
		{
			if (_senderNick.Length == 0)
				return String.Format("Join: {0}", _channel);
			else
				return String.Format("Join: {0} - {1}", _senderNick, _channel);
		}
	}
	public class
	PartMessage : IRCMessage
	{
		private String _channel;
		private String _message;
		
		public String Channel
		{
			get { return _channel; }
			set { _channel = value; }
		}
		
		public
		PartMessage() : base()
		{
			_command = "PART";
		}
		public
		PartMessage(String raw) : base(raw)
		{
			_channel = _commandParams[0];
			_message = _commandParams[1];
			_command = "PART";
		}

		public override String
		ToString()
		{
			if (_senderNick.Length == 0)
				return String.Format("Part: {0} ({1})", _channel, _message);
			else
				return String.Format("Part: {0} - {1} ({2})", _senderNick, _channel, _message);
		}
	}
	public class
	QuitMessage : IRCMessage
	{
		private String _message;

		public
		QuitMessage() : base()
		{
			_command = "QUIT";
		}
		public
		QuitMessage(String raw) : base(raw)
		{
			_message = _commandParams[0];
			_command = "QUIT";
		}
		public
		QuitMessage(String msg, String param)
		{
			_command = "QUIT";
			_commandParams[0] = _message = msg;
		}

		public override String
		ToString()
		{
			return String.Format("Quit: {0} ({1})", _senderNick, _message);
		}
	}
	public class
	TopicMessage : IRCMessage
	{
		private String _topic;
		private String _channel;

		public
		TopicMessage() : base()
		{
			_command = "TOPIC";
		}
		public
		TopicMessage(String raw) : base(raw)
		{
			_channel = _commandParams[0];
			_topic = _commandParams[1];
			_command = "TOPIC";
		}
		public
		TopicMessage(String channel, String param)
		{
			_command = "TOPIC";
			_commandParams[0] = _channel = channel;
			_commandParams[1] = _topic = param;
		}

		public override String
		ToString()
		{
			return String.Format("Topic: [{0}] {1}", _channel, _topic);
		}
	}	
	public class IRCSessionException : ApplicationException
	{
		public IRCSessionException(String message) : base(message) {}
	}
	public class IRCNotConnectedException : IRCSessionException
	{
		public IRCNotConnectedException(String message) : base(message) {}
	}
	public class IRCDisconnectedException : IRCSessionException
	{
		public IRCDisconnectedException(String message) : base(message) {}
	}
	public class NickAlreadyUsedException : IRCSessionException
	{
		public NickAlreadyUsedException(String message) : base(message) {}
	}
	public class PasswordIncorrectException : IRCSessionException
	{
		public PasswordIncorrectException(String message) : base(message) {}
	}
	public class IRCMessageInvalidException : IRCSessionException
	{
		public IRCMessageInvalidException(String message) : base(message) {}
	}
}